#ifndef CHESSY_SEARCH_H
#define CHESSY_SEARCH_H

#include "ChessGame.h"
#include <utility>

namespace chessy
{
    template<typename T>
    class Search;

    struct MoveInfo;

    class MinimaxN;
}


struct chessy::MoveInfo
{
    Move move;
    PieceType movedPiece;
};


template<typename T>
class chessy::Search
{
    T& handler;
    ChessGame& game;
    Board& board;
    Bitboard whites;
    Bitboard blacks;
public:
    inline Search(T& handler, ChessGame& game) :
        handler{ handler }, game{ game }, board{ game.getBoard() },
        whites{ board.getWhites() }, blacks{ board.getBlacks() } {}


    template<Side side, typename... Args>
    inline void iterateAll(Args&&... args);

    template<Side side, typename... Args>
    inline void iterateKnights(Args&&... args);

    template<Side side, typename... Args>
    inline void iterateBishops(Args&&... args);

    template<Side side, typename... Args>
    inline void iterateRooks(Args&&... args);

    template<Side side, typename... Args>
    inline void iterateQueens(Args&&... args);

    template<Side side, typename... Args>
    inline void iterateKing(Args&&... args);

    template<Side side, typename... Args>
    inline void iteratePawns(Args&&... args);

    template<Side side, typename... Args>
    inline void iterateSinglePawnPushes(Args&&... args);

    template<Side side, typename... Args>
    inline void iterateDoublePawnPushes(Args&&... args);

    template<Side side, typename... Args>
    inline void iteratePawnCaptures(Args&&... args);

    template<Side side, typename... Args>
    inline void iteratePromotions(Args&&... args);
private:

    template<Side side>
    inline Bitboard friends(void) const
    {
        return side == WHITE_SIDE ? whites : blacks;
    }

    template<Side side>
    inline Bitboard enemies(void) const
    {
        return side == WHITE_SIDE ? blacks : whites;
    }
};


class chessy::MinimaxN
{
public:
    bool operator () (int a, int b) {
        return true;
    }
};


/*inline void aga() {
    chessy::ChessGame cg;
    chessy::MinimaxN mm;
    chessy::Search<chessy::MinimaxN> search = { mm, cg };
    search.iterateAll(5, 7);
}*/


template<typename T>
template<chessy::Side side, typename... Args>
inline void chessy::Search<T>::iterateAll(Args&&... args)
{
    iteratePawns<side, Args...>(std::forward<Args>(args)...);
    iterateKnights<side, Args...>(std::forward<Args>(args)...);
    iterateBishops<side, Args...>(std::forward<Args>(args)...);
    iterateRooks<side, Args...>(std::forward<Args>(args)...);
    iterateQueens<side, Args...>(std::forward<Args>(args)...);
    iterateKing<side, Args...>(std::forward<Args>(args)...);

    /*


    Bitboard& rs = board.getRooks<side>();
    PositionSet rooks { rs }; 
    for (auto rook : rooks) {
        for (auto pos : PrimitiveRookMoveGenerator{ rook, enemies, friends }) {
            Move move = { rook, pos };
            board.removeAt(move.destination);
            rs.applyMove(move);
            //BestMove m = minimax<otherSide(side)>(depth - 1);
            m.move = move;
            bestMove.overwriteIfBetter(m);
            board = temp;
        }
    }

    Bitboard& qs = board.getQueens<side>();
    PositionSet queens { qs };
    for (auto queen : queens) {
        for (auto pos : PrimitiveQueenMoveGenerator{ queen, enemies, friends }) {
            Move move = { queen, pos };
            board.removeAt(move.destination);
            qs.applyMove(move);
            //BestMove m = minimax<otherSide(side)>(depth - 1);
            m.move = move;
            bestMove.overwriteIfBetter(m);
            board = temp;
        }
    }

    Bitboard& king = board.getKing<side>();
    Index kingIndex = king.getLeastSignificantBit();
    for (auto pos : KingMoveGenerator{ king, friends }) {
        Move move = { kingIndex, pos };
        board.removeAt(pos);
        king.applyMove(move);
        BestMove m = minimax<otherSide(side)>(depth - 1);
        m.move = move;
        //if (depth >= 3)
        //    std::cout << m.move.asString() << " " << bestMove.value << " -> " << m.value << std::endl;
        bestMove.overwriteIfBetter(m);
        board = temp;
    }

    if (side == WHITE_SIDE) {
        if (game.getCanCastleKingSide(side)) {
            if((friends.bits & 0x6) == 0) {
                Move kingMove = {3, 1};
                Move rookMove = {0, 2};
                king.applyMove(kingMove);
                rs.applyMove(rookMove);
                board = temp;
            }
        }
    }

    float v = evaluate<side>();
    bestMove.value *= 0.9f;
    bestMove.value += v * 0.2f;
    return -bestMove;*/
}


template<typename T>
template<chessy::Side side, typename... Args>
inline void chessy::Search<T>::iterateKnights(Args&&... args)
{
    MoveInfo moveInfo;
    Bitboard& ns = board.getKnights<side>();
    PositionSet knights{ ns };
    moveInfo.movedPiece = PieceType::KNIGHT;
    for (auto knight : knights) {
        for (auto pos : KnightMoveGenerator{ knight, friends<side>() }) {
            moveInfo.move = Move{ knight, pos };
            handler(moveInfo, std::forward<Args>(args)...);
        }
    }
}


template<typename T>
template<chessy::Side side, typename... Args>
inline void chessy::Search<T>::iterateBishops(Args&&... args)
{
    MoveInfo moveInfo;
    moveInfo.movedPiece = PieceType::BISHOP;
    PositionSet bishops { board.getBishops<side>() }; 
    for (auto bishop : bishops) {
        for (auto pos : PrimitiveBishopMoveGenerator{
            bishop, enemies<side>(), friends<side>() }) {
            moveInfo.move = Move{ bishop, pos };
            handler(moveInfo, std::forward<Args>(args)...);
        }
    }
}


template<typename T>
template<chessy::Side side, typename... Args>
inline void chessy::Search<T>::iterateRooks(Args&&... args)
{
    MoveInfo moveInfo;
    moveInfo.movedPiece = PieceType::ROOK;
    PositionSet rooks { board.getRooks<side>() }; 
    for (auto rook : rooks) {
        for (auto pos : PrimitiveRookMoveGenerator{
            rook, enemies<side>(), friends<side>() }) {

            moveInfo.move = Move{ rook, pos };
            handler(moveInfo, std::forward<Args>(args)...);
        }
    }
}


template<typename T>
template<chessy::Side side, typename... Args>
inline void chessy::Search<T>::iterateQueens(Args&&... args)
{
    MoveInfo moveInfo;
    moveInfo.movedPiece = PieceType::QUEEN;
    PositionSet queens { board.getQueens<side>() }; 
    for (auto queen : queens) {
        for (auto pos : PrimitiveQueenMoveGenerator{
            queen, enemies<side>(), friends<side>() }) {

            moveInfo.move = Move{ queen, pos };
            handler(moveInfo, std::forward<Args>(args)...);
        }
    }
}


template<typename T>
template<chessy::Side side, typename... Args>
inline void chessy::Search<T>::iterateKing(Args&&... args)
{
    MoveInfo moveInfo;
    moveInfo.movedPiece = PieceType::KING;
    Bitboard king = board.getKing<side>();
    for (auto pos : KingMoveGenerator{ king, friends<side>() }) {
        moveInfo.move = Move{ king.getLeastSignificantBit(), pos };
        handler(moveInfo, std::forward<Args>(args)...);
    }
}


template<typename T>
template<chessy::Side side, typename... Args>
inline void chessy::Search<T>::iterateSinglePawnPushes(Args&&... args)
{
    MoveInfo moveInfo;
    moveInfo.movedPiece = PieceType::PAWN;
    PawnPushGenerator<side> mg{ game };
    for (Move push : mg) {
        moveInfo.move = push;
        handler(moveInfo, std::forward<Args>(args)...);
    }
}


template<typename T>
template<chessy::Side side, typename... Args>
inline void chessy::Search<T>::iterateDoublePawnPushes(Args&&... args)
{
    MoveInfo moveInfo;
    moveInfo.movedPiece = PieceType::PAWN;
    PawnDoublePushGenerator<side> mg{ game };
    for (Move push : mg) {
        moveInfo.move = push;
        handler(moveInfo, std::forward<Args>(args)...);
    }
}


template<typename T>
template<chessy::Side side, typename... Args>
inline void chessy::Search<T>::iteratePawnCaptures(Args&&... args)
{
    MoveInfo moveInfo;
    moveInfo.movedPiece = PieceType::PAWN;
    PawnCaptureGenerator<side, LEFT> mgl{ game };
    for (Move push : mgl) {
        moveInfo.move = push;
        handler(moveInfo, std::forward<Args>(args)...);
    }
    PawnCaptureGenerator<side, RIGHT> mgr{ game };
    for (Move push : mgr) {
        moveInfo.move = push;
        handler(moveInfo, std::forward<Args>(args)...);
    }
}


template<typename T>
template<chessy::Side side, typename... Args>
inline void chessy::Search<T>::iteratePromotions(Args&&... args)
{
    MoveInfo moveInfo;
    moveInfo.movedPiece = PieceType::PAWN;
    PromotionGenerator<side> pg{ game };
    for (Move promotion : pg) {
        moveInfo.move = promotion;
        handler(moveInfo, std::forward<Args>(args)...);
    }
}


template<typename T>
template<chessy::Side side, typename... Args>
inline void chessy::Search<T>::iteratePawns(Args&&... args)
{
    iterateSinglePawnPushes<side, Args...>(std::forward(args)...);
    iterateDoublePawnPushes<side, Args...>(std::forward(args)...);
    iteratePawnCaptures<side, Args...>(std::forward(args)...);
    iteratePromotions<side, Args...>(std::forward(args)...);
    /*
    PawnCaptureGenerator<side, LEFT> pl{ game };
    for (Move capture : pl) {
        Bitboard& pawns = board.getPawns<side>();
        board.removeAt(capture.destination);
        pawns.applyMove(capture);
        BestMove m = minimax<otherSide(side)>(depth - 1);
        m.move = capture;
        bestMove.overwriteIfBetter(m);
        board = temp;
    }

    PawnCaptureGenerator<side, RIGHT> pr{ game };
    for (Move capture : pr) {
        Bitboard& pawns = board.getPawns<side>();
        board.removeAt(capture.destination);
        pawns.applyMove(capture);
        BestMove m = minimax<otherSide(side)>(depth - 1);
        m.move = capture;
        bestMove.overwriteIfBetter(m);
        board = temp;
    }

    PromotionGenerator<side> pg{ game };
    for (Move promotion : pg) {
        Bitboard& pawns = board.getPawns<side>();
        board.removeAt(promotion.destination);
        pawns.applyMove(promotion);
        BestMove m = minimax<otherSide(side)>(depth - 1);
        m.move = promotion;
        bestMove.overwriteIfBetter(m);
        board = temp;
    }*/
}


#endif // CHESSY_SEARCH_H
